# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
from typing import Union, List, Optional, Any, Dict, Tuple

import yaml

from kubemarine import system
from kubemarine.core import utils

from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.group import RunnersGroupResult


def proceed_section_keyvalue(data: Dict[str, Any], tabsize: int) -> str:
    tab = " "*tabsize
    config = ''

    for key, value in data.items():
        if isinstance(value, bool):
            if value:
                config += '\n' + tab + '%s' % key
            continue
        if isinstance(value, str) or isinstance(value, int) and value:
            if not isinstance(value, int) and any((c in set(' ')) for c in value):
                config += '\n' + tab + '%s \"%s\"' % (key, value)
            elif isinstance(value, str) and len(value) == 0:
                config += '\n' + tab + '%s' % key
            else:
                config += '\n' + tab + '%s %s' % (key, value)
            continue
        if isinstance(value, list) and value:
            config += '\n' + tab + '%s %s' % (key, " ".join(value))
            continue
        if isinstance(value, dict):
            config += generate_nested_sections(key, value, tabsize)
            continue
        raise Exception('Unknown type of field in coredns services')

    return config


def generate_nested_sections(type_: str, data: Dict[str, Dict[str, Any]], tabsize: int) -> str:
    tab = " "*tabsize
    config = ''

    max_priority = 0
    for section_name, section_value in data.items():
        if section_value.get('priority') is not None and section_value['priority'] > max_priority:
            max_priority = section_value['priority']

    iterated = 0
    sections: List[Tuple[str, int]] = []
    for section_name, section_value in data.items():
        if section_value.get('priority') is None:
            iterated += 1
            section_priority = max_priority + iterated
        else:
            section_priority = section_value['priority']

        if section_value.get('enabled', True) in ['1', 1, True, 'True']:
            sections.append((section_name, section_priority))

    sections = sorted(sections, key=lambda i: i[1])

    for section in sections:
        section_name, _ = section
        section_value = data[section_name]
        if type_ == 'kubernetes':
            config += '\n' + tab + type_
            if section_value.get('zone'):
                if isinstance(section_value['zone'], list):
                    section_value['zone'] = ' '.join(section_value['zone'])
                config += ' ' + section_value['zone']
            config += ' {' + proceed_section_keyvalue(section_value['data'], tabsize + 2) + '\n' + tab + '}'

        elif type_ == 'hosts':
            config += '\n' + tab + type_
            if section_value.get('file') and isinstance(section_value['file'], str):
                config += ' ' + section_value['file']
            config += ' {' + proceed_section_keyvalue(section_value['data'], tabsize + 2) + '\n' + tab + '}'

        elif type_ == 'template':
            zones: Union[str, List[Optional[str]]] = [None]
            if section_value.get('zone'):
                zones = section_value['zone']
                if isinstance(zones, str):
                    zones = [zones]
            for zone in zones:
                config += '\n' + tab + type_
                if section_value.get('class'):
                    config += ' ' + section_value['class']
                if section_value.get('type'):
                    config += ' ' + section_value['type']
                if zone:
                    config += ' ' + zone
                config += ' {' + proceed_section_keyvalue(section_value['data'], tabsize + 2) + '\n' + tab + '}'

        else:
            config += '\n' + tab + type_ + ' {' + proceed_section_keyvalue(section_value['data'], tabsize + 2) \
                      + '\n' + tab + '}'

    return config


def generate_configmap(inventory: dict) -> str:
    # coredns.configmap.Hosts must exist even if it's empty
    if not inventory['services']['coredns']['configmap'].get('Hosts'):
        # Hosts must exist even if it's empty
        inventory['services']['coredns']['configmap']['Hosts'] = ""

    config = '''apiVersion: v1

kind: ConfigMap
metadata:
  name: coredns
  namespace: kube-system
data:'''

    for config_type, data in inventory['services']['coredns']['configmap'].items():
        config += '\n  %s: |' % config_type
        if config_type == 'Corefile':
            for port, settings in data.items():
                config += '\n    %s {' % port
                config += proceed_section_keyvalue(settings, 6)
                config += '\n    }'
        else:
            config += '\n    ' + data.replace('\n', '\n    ')
        # add etc_hosts_autogenerated to the configmap
        if config_type == 'Hosts' and inventory['services']['coredns']['add_etc_hosts_generated']:
            config += '\n    '
            config += system.generate_etc_hosts_config(inventory,etc_hosts_part='etc_hosts_generated').replace('\n', '\n    ')

    return config + '\n'


def apply_configmap(cluster: KubernetesCluster, config: str) -> RunnersGroupResult:
    utils.dump_file(cluster, config, 'coredns-configmap.yaml')

    group = cluster.make_group_from_roles(['control-plane'])
    group.put(io.StringIO(config), '/etc/kubernetes/coredns-configmap.yaml', backup=True, sudo=True)

    return cluster.nodes['control-plane'].get_first_member()\
        .sudo('kubectl apply -f /etc/kubernetes/coredns-configmap.yaml && '
             'sudo kubectl rollout restart -n kube-system deployment/coredns')


def apply_patch(cluster: KubernetesCluster) -> Union[RunnersGroupResult, str]:
    apply_command = ''

    for config_type in ['deployment']:

        if not cluster.inventory['services']['coredns'].get(config_type):
            continue

        if apply_command != '':
            apply_command += ' && sudo '

        config = yaml.dump(cluster.inventory['services']['coredns'][config_type])
        filename = 'coredns-%s-patch.yaml' % config_type
        filepath = '/etc/kubernetes/' + filename

        utils.dump_file(cluster, config, filename)

        group = cluster.make_group_from_roles(['control-plane'])
        group.put(io.StringIO(config), filepath, backup=True, sudo=True)

        apply_command = 'kubectl patch %s coredns -n kube-system --type merge -p \"$(sudo cat %s)\"' % (config_type, filepath)

    if apply_command == '':
        return 'Nothing to patch'

    return cluster.nodes['control-plane'].get_first_member().sudo(apply_command)



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from textwrap import dedent
from typing import List, Optional, Dict
import yaml

from kubemarine.core import summary, log
from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage, enrichment
from kubemarine.plugins.manifest import Processor, EnrichmentFunction, Manifest, Identity


@enrichment(EnrichmentStage.FULL)
def enrich_inventory(cluster: KubernetesCluster) -> None:
    inventory = cluster.inventory
    if not inventory["plugins"]["kubernetes-dashboard"]["install"]:
        return

    # if user defined resources himself, we should use them as is, instead of merging with our defaults
    raw_dashboard = cluster.raw_inventory.get("plugins", {}).get("kubernetes-dashboard", {}).get("dashboard", {})
    if "resources" in raw_dashboard:
        inventory["plugins"]["kubernetes-dashboard"]["dashboard"]["resources"] = raw_dashboard["resources"]
    raw_metrics_scrapper = cluster.raw_inventory.get("plugins", {}).get("kubernetes-dashboard", {}).get("metrics-scraper", {})
    if "resources" in raw_metrics_scrapper:
        inventory["plugins"]["kubernetes-dashboard"]["metrics-scraper"]["resources"] = raw_metrics_scrapper["resources"]


def schedule_summary_report(cluster: KubernetesCluster) -> None:
    plugin_item = cluster.inventory['plugins']['kubernetes-dashboard']
    hostname = plugin_item['hostname']
    # Currently we declare that Dashboard UI is available only via HTTPS
    summary.schedule_report(cluster.context, summary.SummaryItem.DASHBOARD_URL, f'https://{hostname}')


class DashboardManifestProcessor(Processor):
    def __init__(self, logger: log.VerboseLogger, inventory: dict,
                 original_yaml_path: Optional[str] = None, destination_name: Optional[str] = None) -> None:
        super().__init__(logger, inventory, Identity('kubernetes-dashboard'), original_yaml_path, destination_name)

    def get_known_objects(self) -> List[str]:
        return [
            "Namespace_kubernetes-dashboard",
            "ServiceAccount_kubernetes-dashboard",
            "Service_kubernetes-dashboard",
            "Secret_kubernetes-dashboard-certs",
            "Secret_kubernetes-dashboard-csrf",
            "Secret_kubernetes-dashboard-key-holder",
            "ConfigMap_kubernetes-dashboard-settings",
            "Role_kubernetes-dashboard",
            "ClusterRole_kubernetes-dashboard",
            "RoleBinding_kubernetes-dashboard",
            "ClusterRoleBinding_kubernetes-dashboard",
            "Deployment_kubernetes-dashboard",
            "Service_dashboard-metrics-scraper",
            "Deployment_dashboard-metrics-scraper",
        ]

    def get_enrichment_functions(self) -> List[EnrichmentFunction]:
        return [
            self.enrich_namespace_kubernetes_dashboard,
            self.enrich_service_account_secret_kubernetes_dashboard,
            self.enrich_service_account_kubernetes_dashboard,
            self.enrich_deployment_kubernetes_dashboard,
            self.enrich_deployment_dashboard_metrics_scraper,
        ]

    def get_namespace_to_necessary_pss_profiles(self) -> Dict[str, str]:
        return {'kubernetes-dashboard': 'baseline'}

    def enrich_namespace_kubernetes_dashboard(self, manifest: Manifest) -> None:
        self.assign_default_pss_labels(manifest, 'kubernetes-dashboard')

    def enrich_service_account_secret_kubernetes_dashboard(self, manifest: Manifest) -> None:
        new_yaml = yaml.safe_load(service_account_secret_kubernetes_dashboard)

        service_account_key = "ServiceAccount_kubernetes-dashboard"
        service_account_index = manifest.all_obj_keys().index(service_account_key) \
            if service_account_key in manifest.all_obj_keys() else -1
        
        self.include(manifest, service_account_index + 1, new_yaml)

    def enrich_service_account_kubernetes_dashboard(self, manifest: Manifest) -> None:
        key = "ServiceAccount_kubernetes-dashboard"
        source_yaml = manifest.get_obj(key, patch=True)
        source_yaml['automountServiceAccountToken'] = False

    def enrich_deployment_kubernetes_dashboard(self, manifest: Manifest) -> None:
        key = "Deployment_kubernetes-dashboard"
        source_yaml = manifest.get_obj(key, patch=True)
        
        service_account_name = "kubernetes-dashboard"
        self.enrich_volume_and_volumemount(source_yaml, service_account_name)
       
        self.log.verbose(f"The {key} has been updated to include the new secret volume and mount.")

        self.enrich_image_for_container(manifest, key,
            plugin_service='dashboard', container_name='kubernetes-dashboard', is_init_container=False)

        self.enrich_resources_for_container(manifest, key, container_name='kubernetes-dashboard', plugin_service="dashboard")
        self.enrich_node_selector(manifest, key, plugin_service='dashboard')
        self.enrich_tolerations(manifest, key, plugin_service='dashboard', override=True)

    def enrich_deployment_dashboard_metrics_scraper(self, manifest: Manifest) -> None:
        key = "Deployment_dashboard-metrics-scraper"
        source_yaml = manifest.get_obj(key, patch=True)
        
        service_account_name = "kubernetes-dashboard"
        self.enrich_volume_and_volumemount(source_yaml, service_account_name)
       
        self.log.verbose(f"The {key} has been updated to include the new secret volume and mount.")

        self.enrich_image_for_container(manifest, key,
            plugin_service='metrics-scraper', container_name='dashboard-metrics-scraper', is_init_container=False)

        self.enrich_resources_for_container(
            manifest, key, container_name='dashboard-metrics-scraper', plugin_service="metrics-scraper")
        self.enrich_node_selector(manifest, key, plugin_service='metrics-scraper')
        self.enrich_tolerations(manifest, key, plugin_service='metrics-scraper', override=True)


service_account_secret_kubernetes_dashboard = dedent("""\
    apiVersion: v1
    kind: Secret
    metadata:
      name: kubernetes-dashboard-token
      namespace: kubernetes-dashboard
      annotations:
        kubernetes.io/service-account.name: kubernetes-dashboard
    type: kubernetes.io/service-account-token  
""")



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Simple TCP socket listener that can be run on both python 2 and 3,
# The listener accepts connections sequentially, and suppresses the received data.
# The script is for testing purpose only.
# The first argv parameter is the TCP port to listen. The second argv parameter is the ip protocol version.

import socket
import sys

port = int(sys.argv[1])

proto = '{{ proto }}'
ip_version = '{{ ip_version }}'

s_type = socket.SOCK_STREAM
if proto == 'udp':
    s_type = socket.SOCK_DGRAM

family = socket.AF_INET
if ip_version == '6':
    family = socket.AF_INET6

s = socket.socket(family, s_type)
try:
    try:
        s.bind(('{{ address }}', port))
    except socket.error as e:
        if "Address already in use" in str(e):
            sys.stdout.write("In use\n")
            sys.stdout.flush()
            sys.exit(1)
        else:
            raise

    sys.stdout.write("Listen\n")
    sys.stdout.flush()

    sz = int('{{ mtu }}')

    if proto == 'udp':
        while True:
            data, address = s.recvfrom(sz)
            s.sendto(data, address)

    else:
        s.listen(1)

        while True:
            client, _ = s.accept()
            try:
                while True:
                    data = client.recv(sz)
                    if not data:
                        break
                    client.sendall(data)
            finally:
                client.close()
finally:
    s.close()



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from textwrap import dedent
from typing import Optional, List, Dict

import os
import yaml

from kubemarine import plugins
from kubemarine.core import utils, log
from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage, enrichment
from kubemarine.core.group import NodeGroup
from kubemarine.kubernetes import secrets
from kubemarine.plugins.manifest import Processor, EnrichmentFunction, Manifest, Identity


@enrichment(EnrichmentStage.FULL)
def enrich_inventory(cluster: KubernetesCluster) -> None:
    inventory = cluster.inventory
    if not inventory["plugins"]["calico"]["install"]:
        return

    # if user defined resources himself, we should use them as is, instead of merging with our defaults
    raw_calico_node = cluster.raw_inventory.get("plugins", {}).get("calico", {}).get("node", {})
    if "resources" in raw_calico_node:
        inventory["plugins"]["calico"]["node"]["resources"] = raw_calico_node["resources"]
    raw_typha = cluster.raw_inventory.get("plugins", {}).get("calico", {}).get("typha", {})
    if "resources" in raw_typha:
        inventory["plugins"]["calico"]["typha"]["resources"] = raw_typha["resources"]
    raw_calico_controller = cluster.raw_inventory.get("plugins", {}).get("calico", {}).get("kube-controllers", {})
    if "resources" in raw_calico_controller:
        inventory["plugins"]["calico"]["kube-controllers"]["resources"] = raw_calico_controller["resources"]
    raw_apiserver = cluster.raw_inventory.get("plugins", {}).get("calico", {}).get("apiserver", {})
    if "resources" in raw_apiserver:
        inventory["plugins"]["calico"]["apiserver"]["resources"] = raw_apiserver["resources"]


@enrichment(EnrichmentStage.PROCEDURE, procedures=['remove_node'])
def enrich_remove_node_set_previous_typha_enabled(cluster: KubernetesCluster) -> None:
    if not cluster.previous_inventory["plugins"]["calico"]["install"]:
        return

    # If Typha was (effectively) enabled before removing of node and `enabled` value is not explicitly provided,
    # Let's set previous flag in all inventories including final, as we are not able to fully remove Typha.
    typha_enabled_before = is_typha_enabled(cluster.previous_inventory)
    if typha_enabled_before and 'enabled' not in cluster.inventory.get('plugins', {}).get('calico', {}).get('typha', {}):
        cluster.inventory.setdefault('plugins', {}).setdefault('calico', {})\
            .setdefault('typha', {})['enabled'] = typha_enabled_before


def new_route_reflectors_added(cluster: KubernetesCluster) -> bool:
    procedure = cluster.context['initial_procedure']
    if procedure != 'add_node':
        return False

    calico = cluster.inventory['plugins']['calico']
    if calico['fullmesh']:
        return False

    for node in cluster.get_new_nodes().having_roles(['control-plane', 'worker']).get_ordered_members_configs_list():
        if node.get('labels', {}).get('route-reflector') is True:
            break
    else:
        return False

    steps = calico['installation']['procedures']
    for step in steps:
        if 'template' not in step:
            continue

        if step['template']['source'] == 'templates/plugins/calico-rr.sh.j2':
            break
    else:
        return False

    return True


# DEPRECATED
def apply_calico_yaml(cluster: KubernetesCluster, calico_original_yaml: str, calico_yaml: str) -> None:
    """
    The method implements full processing for Calico plugin
    :param calico_original_yaml: path to original Calico manifest
    :param calico_yaml: file name of the resulting Calico manifest
    :param cluster: Cluster object
    """

    calico_yaml = os.path.basename(calico_yaml)
    processor = CalicoManifestProcessor(cluster.log, cluster.inventory,
                                        original_yaml_path=calico_original_yaml,
                                        destination_name=calico_yaml)
    manifest = processor.enrich()
    processor.apply(cluster, manifest)


def is_typha_enabled(inventory: dict) -> bool:
    enabled: bool = inventory['plugins']['calico']['typha']['enabled']
    return enabled


def is_apiserver_enabled(inventory: dict) -> bool:
    enabled: bool = inventory['plugins']['calico']['apiserver']['enabled']
    return enabled


def renew_apiserver_certificate(cluster: KubernetesCluster) -> None:
    from kubemarine import kubernetes  # pylint: disable=cyclic-import

    logger = cluster.log
    if not is_apiserver_enabled(cluster.inventory):
        logger.debug("Calico API server is disabled. Skip renewing of the key and certificate.")
        return

    namespace = "calico-apiserver"
    secret_name = "calico-apiserver-certs"
    deployment = "calico-apiserver"

    control_planes = cluster.nodes["control-plane"]
    first_control_plane = control_planes.get_first_member()

    with secrets.create_tls_secret_procedure(first_control_plane):
        logger.debug("Creating or renewing of the key and certificate for the Calico API server")

        config = dedent(
            """\
            [req]
            distinguished_name = req
            [v3_req]
            basicConstraints = critical,CA:TRUE
            subjectAltName = DNS:calico-api.calico-apiserver.svc
            """
        )
        first_control_plane.call(
            secrets.create_certificate,
            config=config,
            customization_flags='-newkey rsa:4096 -days 365 -subj "/" -extensions v3_req')

        first_control_plane.call(secrets.renew_tls_secret, name=secret_name, namespace=namespace)

        logger.debug("Patching the APIService for the Calico API server")

        first_control_plane.sudo(
            f"{secrets.get_encoded_certificate_cmd()} "
            "| xargs -I CERT "
            "sudo kubectl patch apiservice v3.projectcalico.org "
            r'-p "{\"spec\": {\"caBundle\": \"CERT\"}}"')

    # Force restart the deployment instead of graceful waiting for the automatic secret propagation.
    # This is necessary to make the procedure independent on the secret propagation timeout.
    logger.debug("Restarting the Calico API server deployment")
    first_control_plane.sudo(f"kubectl rollout restart -n {namespace} deployment {deployment}")
    plugins.expect_deployment(cluster, [{'name': deployment, 'namespace': namespace}])
    plugins.expect_pods(cluster, ['calico-apiserver'], namespace=namespace)

    logger.debug("Waiting for the Calico API service availability through the Kubernetes API server")

    # Try to access some projectcalico.org resource using each instance of the Kubernetes API server.
    expect_config = cluster.inventory['plugins']['calico']['apiserver']['expect']['apiservice']
    with kubernetes.local_admin_config(control_planes) as kubeconfig:
        control_planes.call(NodeGroup.wait_command_successful,
                            command=f"kubectl --kubeconfig {kubeconfig} get ippools.projectcalico.org > /dev/null",
                            hide=True, pty=True,
                            retries=expect_config['retries'], timeout=expect_config['timeout'])


class CalicoManifestProcessor(Processor):
    def __init__(self, logger: log.VerboseLogger, inventory: dict,
                 original_yaml_path: Optional[str] = None, destination_name: Optional[str] = None):
        super().__init__(logger, inventory, Identity('calico'), original_yaml_path, destination_name)

    def exclude_typha_objects_if_disabled(self, manifest: Manifest) -> None:
        # enrich 'calico-typha' objects only if it's enabled in 'cluster.yaml'
        # in other case those objects must be excluded
        if not is_typha_enabled(self.inventory):
            for key in ("Deployment_calico-typha", "Service_calico-typha", "PodDisruptionBudget_calico-typha"):
                self.exclude(manifest, key)

    def enrich_configmap_calico_config(self, manifest: Manifest) -> None:
        """
        The method implements the enrichment procedure for Calico ConfigMap
        :param manifest: Container to operate with manifest objects
        """

        key = "ConfigMap_calico-config"
        source_yaml = manifest.get_obj(key, patch=True)
        val = self.inventory['plugins']['calico']['mtu']
        source_yaml['data']['veth_mtu'] = str(val)
        self.log.verbose(f"The {key} has been patched in 'data.veth_mtu' with '{val}'")
        val = "calico-typha" if is_typha_enabled(self.inventory) else "none"
        source_yaml['data']['typha_service_name'] = val
        self.log.verbose(f"The {key} has been patched in 'data.typha_service_name' with '{val}'")
        string_part = source_yaml['data']['cni_network_config']
        ip = self.inventory['services']['kubeadm']['networking']['podSubnet'].split('/')[0]
        if utils.isipv(ip, [4]):
            val = self.inventory['plugins']['calico']['cni']['ipam']['ipv4']
        else:
            val = self.inventory['plugins']['calico']['cni']['ipam']['ipv6']
        new_string_part = string_part.replace('"type": "calico-ipam"', str(val)[:-1][1:].replace("'", "\""))
        source_yaml['data']['cni_network_config'] = new_string_part
        log_str = new_string_part.replace("\n", "")
        self.log.verbose(f"The {key} has been patched in 'data.cni_network_config' with '{log_str}'")

    def enrich_service_account_secret_calico_kube_controllers(self, manifest: Manifest) -> None:
        new_yaml = yaml.safe_load(service_account_secret_calico_kube_controllers)

        service_account_key = "ServiceAccount_calico-kube-controllers"
        service_account_index = manifest.all_obj_keys().index(service_account_key) \
            if service_account_key in manifest.all_obj_keys() else -1
        
        self.include(manifest, service_account_index + 1, new_yaml)

    def enrich_service_account_calico_kube_controllers(self, manifest: Manifest) -> None:
        key = "ServiceAccount_calico-kube-controllers"
        source_yaml = manifest.get_obj(key, patch=True)
        source_yaml['automountServiceAccountToken'] = False
    
    def enrich_deployment_calico_kube_controllers(self, manifest: Manifest) -> None:
        """
        The method implements the enrichment procedure for Calico controller Deployment
        :param manifest: Container to operate with manifest objects
        """

        key = "Deployment_calico-kube-controllers"
        service_account_name = "calico-kube-controllers"
        source_yaml= manifest.get_obj(key, patch=True)
        self.enrich_volume_and_volumemount(source_yaml, service_account_name)
        self.enrich_node_selector(manifest, key, plugin_service='kube-controllers')
        self.enrich_tolerations(manifest, key, plugin_service='kube-controllers')
        self.enrich_resources_for_container(manifest, key,
            plugin_service='kube-controllers', container_name='calico-kube-controllers')
        self.enrich_image_for_container(manifest, key,
            plugin_service='kube-controllers', container_name='calico-kube-controllers', is_init_container=False)

    def enrich_service_account_secret_calico_node(self, manifest: Manifest) -> None:
        new_yaml = yaml.safe_load(service_account_secret_calico_node)

        service_account_key = "ServiceAccount_calico-node"
        service_account_index = manifest.all_obj_keys().index(service_account_key) \
            if service_account_key in manifest.all_obj_keys() else -1
        
        self.include(manifest, service_account_index + 1, new_yaml)

    def enrich_service_account_calico_node(self, manifest: Manifest) -> None:
        key = "ServiceAccount_calico-node"
        source_yaml = manifest.get_obj(key, patch=True)
        source_yaml['automountServiceAccountToken'] = False
    
    def enrich_daemonset_calico_node(self, manifest: Manifest) -> None:
        """
        The method implements the enrichment procedure for Calico node DaemonSet
        :param manifest: Container to operate with manifest objects
        """

        key = "DaemonSet_calico-node"

        service_account_name = "calico-node"
        init_container_name = "upgrade-ipam"
        source_yaml= manifest.get_obj(key, patch=True)
        self.enrich_volume_and_volumemount(source_yaml, service_account_name, init_container_name)

        for container_name in ['upgrade-ipam', 'install-cni']:
            self.enrich_image_for_container(manifest, key,
                plugin_service='cni', container_name=container_name, is_init_container=True)

        self.enrich_image_for_container(manifest, key,
            plugin_service='node', container_name='mount-bpffs', is_init_container=True, allow_absent=True)
        self.enrich_image_for_container(manifest, key,
            plugin_service='flexvol', container_name='flexvol-driver', is_init_container=True, allow_absent=True)

        self.enrich_image_for_container(manifest, key,
            plugin_service='node', container_name='calico-node', is_init_container=False)

        self.enrich_daemonset_calico_node_container_env(manifest)
        self.enrich_resources_for_container(manifest, key,
            plugin_service='node', container_name='calico-node')

    def enrich_daemonset_calico_node_container_env(self, manifest: Manifest) -> None:
        """
        The method implements the enrichment procedure for 'calico-node' container in Calico node DaemonSet.
        The method attempts to preserve initial formatting.

        :param manifest: Container to operate with manifest objects
        """
        key = "DaemonSet_calico-node"
        env_delete: List[str] = []
        ip = self.inventory['services']['kubeadm']['networking']['podSubnet'].split('/')[0]
        if utils.isipv(ip, [4]):
            env_delete.extend([
                'CALICO_IPV6POOL_CIDR', 'IP6', 'IP6_AUTODETECTION_METHOD',
                'CALICO_IPV6POOL_IPIP', 'CALICO_IPV6POOL_VXLAN'
            ])
        if not is_typha_enabled(self.inventory):
            env_delete.append('FELIX_TYPHAK8SSERVICENAME')

        env_ensure: Dict[str, str] = {
            # If metrics ports are ever configurable,
            # it makes sense to make it configurable for typha and kube-controllers as well.
            # Metrics services should also be patched.
            'FELIX_PROMETHEUSMETRICSPORT': '9091'
        }

        self.enrich_env_for_container(
            manifest, key, container_name='calico-node', env_delete=env_delete, env_ensure=env_ensure)

    def enrich_deployment_calico_typha(self, manifest: Manifest) -> None:
        """
        The method implements the enrichment procedure for Typha Deployment
        :param manifest: Container to operate with manifest objects
        """

        key = "Deployment_calico-typha"
        if not manifest.has_obj(key):
            return None
        source_yaml = manifest.get_obj(key, patch=True)

        service_account_name = "calico-node"
        self.enrich_volume_and_volumemount(source_yaml, service_account_name)
        self.log.verbose(f"The {key} has been updated to include the new secret volume and mount.")

        default_tolerations = [{'key': 'node.kubernetes.io/network-unavailable', 'effect': 'NoSchedule'},
                               {'key': 'node.kubernetes.io/network-unavailable', 'effect': 'NoExecute'}]

        val = self.inventory['plugins']['calico']['typha']['replicas']
        source_yaml['spec']['replicas'] = val
        self.log.verbose(f"The {key} has been patched in 'spec.replicas' with '{val}'")

        self.enrich_node_selector(manifest, key, plugin_service='typha')
        self.enrich_deployment_calico_typha_tolerations(manifest, default_tolerations)
        self.enrich_image_for_container(manifest, key,
            plugin_service='typha', container_name='calico-typha', is_init_container=False)
        self.enrich_deployment_calico_typha_container_env(manifest)
        self.enrich_resources_for_container(manifest, key,
            plugin_service='typha', container_name='calico-typha')

    def enrich_deployment_calico_typha_tolerations(self, manifest: Manifest, default_tolerations: List[dict]) -> None:
        key = "Deployment_calico-typha"
        self.enrich_tolerations(manifest, key, plugin_service='typha', extra_tolerations=default_tolerations)

    def enrich_deployment_calico_typha_container_env(self, manifest: Manifest) -> None:
        key = "Deployment_calico-typha"
        env_ensure: Dict[str, str] = {
            # If metrics ports are ever configurable, it will need to introduce new `typha.env` section.
            # Also, it makes sense to make it configurable for calico-node and kube-controllers as well.
            # Metrics services should also be patched.
            'TYPHA_PROMETHEUSMETRICSENABLED': 'true',
            'TYPHA_PROMETHEUSMETRICSPORT': '9093'
        }
        # This also searches in calico.typha.env but it is currently not supported by JSON schema.
        self.enrich_env_for_container(
            manifest, key, plugin_service='typha', container_name='calico-typha', env_ensure=env_ensure)

    def enrich_crd_felix_configuration(self, manifest: Manifest) -> None:
        """
        The method implements the enrichment procedure for Calico CRD Felixconfigurations
        :param manifest: Container to operate with manifest objects
        """

        key = "CustomResourceDefinition_felixconfigurations.crd.projectcalico.org"
        source_yaml = manifest.get_obj(key, patch=True)

        api_list = \
        source_yaml['spec']['versions'][0]['schema']['openAPIV3Schema']['properties']['spec']['properties'][
            'prometheusMetricsEnabled']
        api_list["default"] = True
        source_yaml['spec']['versions'][0]['schema']['openAPIV3Schema']['properties']['spec']['properties'][
            'prometheusMetricsEnabled'] = api_list

    def enrich_metrics(self, manifest: Manifest) -> None:
        sz = len(manifest.all_obj_keys())
        yaml = utils.yaml_structure_preserver()
        self.include(manifest, sz, yaml.load(utils.read_internal('templates/plugins/calico-kube-controllers-metrics.yaml')))
        self.include(manifest, sz, yaml.load(utils.read_internal('templates/plugins/calico-metrics.yaml')))
        if is_typha_enabled(self.inventory):
            self.include(manifest, sz, yaml.load(utils.read_internal('templates/plugins/calico-typha-metrics.yaml')))

    def get_known_objects(self) -> List[str]:
        return [
            "ConfigMap_calico-config",
            "CustomResourceDefinition_bgpconfigurations.crd.projectcalico.org",
            "CustomResourceDefinition_bgpfilters.crd.projectcalico.org",
            "CustomResourceDefinition_bgppeers.crd.projectcalico.org",
            "CustomResourceDefinition_blockaffinities.crd.projectcalico.org",
            "CustomResourceDefinition_caliconodestatuses.crd.projectcalico.org",
            "CustomResourceDefinition_clusterinformations.crd.projectcalico.org",
            "CustomResourceDefinition_felixconfigurations.crd.projectcalico.org",
            "CustomResourceDefinition_globalnetworkpolicies.crd.projectcalico.org",
            "CustomResourceDefinition_globalnetworksets.crd.projectcalico.org",
            "CustomResourceDefinition_hostendpoints.crd.projectcalico.org",
            "CustomResourceDefinition_ipamblocks.crd.projectcalico.org",
            "CustomResourceDefinition_ipamconfigs.crd.projectcalico.org",
            "CustomResourceDefinition_ipamhandles.crd.projectcalico.org",
            "CustomResourceDefinition_ippools.crd.projectcalico.org",
            "CustomResourceDefinition_ipreservations.crd.projectcalico.org",
            "CustomResourceDefinition_kubecontrollersconfigurations.crd.projectcalico.org",
            "CustomResourceDefinition_networkpolicies.crd.projectcalico.org",
            "CustomResourceDefinition_networksets.crd.projectcalico.org",
            "ClusterRole_calico-kube-controllers",
            "ClusterRoleBinding_calico-kube-controllers",
            "ClusterRole_calico-node",
            "ClusterRole_calico-cni-plugin",
            "ClusterRoleBinding_calico-node",
            "ClusterRoleBinding_calico-cni-plugin",
            "DaemonSet_calico-node",
            "ServiceAccount_calico-node",
            "ServiceAccount_calico-cni-plugin",
            "Deployment_calico-kube-controllers",
            "ServiceAccount_calico-kube-controllers",
            "PodDisruptionBudget_calico-kube-controllers",
            "Deployment_calico-typha",
            "Service_calico-typha",
            "PodDisruptionBudget_calico-typha"
        ]

    def get_enrichment_functions(self) -> List[EnrichmentFunction]:
        return [
            self.exclude_typha_objects_if_disabled,
            self.enrich_configmap_calico_config,
            self.enrich_service_account_secret_calico_kube_controllers,
            self.enrich_service_account_calico_kube_controllers,
            self.enrich_deployment_calico_kube_controllers,
            self.enrich_service_account_secret_calico_node,
            self.enrich_service_account_calico_node,
            self.enrich_daemonset_calico_node,
            self.enrich_deployment_calico_typha,
            self.enrich_crd_felix_configuration,
            self.enrich_metrics,
        ]


class CalicoLess_3_27_ManifestProcessor(CalicoManifestProcessor):
    def enrich_deployment_calico_typha_tolerations(self, manifest: Manifest, default_tolerations: List[dict]) -> None:
        # "backport" issue https://github.com/projectcalico/calico/pull/7979
        # to allow Typha to be scheduled on control-planes
        for toleration in self.get_typha_schedule_control_plane_extra_tolerations(manifest):
            default_tolerations.insert(0, toleration)

        super().enrich_deployment_calico_typha_tolerations(manifest, default_tolerations)

    def get_typha_schedule_control_plane_extra_tolerations(self, manifest: Manifest) -> List[dict]:
        key = "Deployment_calico-typha"
        source_yaml = manifest.get_obj(key, patch=True)

        schedule_control_plane = [{'effect': 'NoExecute', 'operator': 'Exists'},
                                  {'effect': 'NoSchedule', 'operator': 'Exists'}]

        original_tolerations: List[dict] = source_yaml['spec']['template']['spec'].get('tolerations', [])
        return [toleration for toleration in schedule_control_plane
                if toleration not in original_tolerations]


service_account_secret_calico_node = dedent("""\
    apiVersion: v1
    kind: Secret
    metadata:
      name: calico-node-token
      namespace: kube-system
      annotations:
        kubernetes.io/service-account.name: calico-node
    type: kubernetes.io/service-account-token  
""")
service_account_secret_calico_kube_controllers = dedent("""\
    apiVersion: v1
    kind: Secret
    metadata:
      name: calico-kube-controllers-token
      namespace: kube-system
      annotations:
        kubernetes.io/service-account.name: calico-kube-controllers
    type: kubernetes.io/service-account-token  
""")


def get_calico_manifest_processor(logger: log.VerboseLogger, inventory: dict,
                                  yaml_path: Optional[str] = None, destination: Optional[str] = None) -> Processor:
    version: str = inventory['plugins']['calico']['version']
    kwargs = {'original_yaml_path': yaml_path, 'destination_name': destination}
    if utils.version_key(version)[0:2] < utils.minor_version_key("v3.27"):
        return CalicoLess_3_27_ManifestProcessor(logger, inventory, **kwargs)

    return CalicoManifestProcessor(logger, inventory, **kwargs)


class CalicoApiServerManifestProcessor(Processor):
    def __init__(self, logger: log.VerboseLogger, inventory: dict,
                 original_yaml_path: Optional[str] = None, destination_name: Optional[str] = None):
        super().__init__(logger, inventory, Identity('calico', 'apiserver'), original_yaml_path, destination_name)

    def get_known_objects(self) -> List[str]:
        return [
            "Namespace_calico-apiserver",
            "NetworkPolicy_allow-apiserver",
            "Service_calico-api",
            "Deployment_calico-apiserver",
            "ServiceAccount_calico-apiserver",
            "APIService_v3.projectcalico.org",
            "ClusterRole_calico-crds",
            "ClusterRole_calico-extension-apiserver-auth-access",
            "ClusterRole_calico-webhook-reader",
            "ClusterRoleBinding_calico-apiserver-access-crds",
            "ClusterRoleBinding_calico-apiserver-delegate-auth",
            "ClusterRoleBinding_calico-apiserver-webhook-reader",
            "ClusterRoleBinding_calico-extension-apiserver-auth-access",
        ]

    def get_enrichment_functions(self) -> List[EnrichmentFunction]:
        return [
            self.enrich_namespace_calico_apiserver,
            self.enrich_service_account_secret_calico_apiserver,
            self.enrich_service_account_calico_apiserver,
            self.enrich_deployment_calico_apiserver,
        ]

    def get_namespace_to_necessary_pss_profiles(self) -> Dict[str, str]:
        return {'calico-apiserver': 'baseline'}

    def enrich_namespace_calico_apiserver(self, manifest: Manifest) -> None:
        self.assign_default_pss_labels(manifest, 'calico-apiserver')

    def enrich_service_account_secret_calico_apiserver(self, manifest: Manifest) -> None:
        new_yaml = yaml.safe_load(service_account_secret_calico_apiserver)

        service_account_key = "ServiceAccount_calico-apiserver"
        service_account_index = manifest.all_obj_keys().index(service_account_key) \
            if service_account_key in manifest.all_obj_keys() else -1
        
        self.include(manifest, service_account_index + 1, new_yaml)

    def enrich_service_account_calico_apiserver(self, manifest: Manifest) -> None:
        key = "ServiceAccount_calico-apiserver"
        source_yaml = manifest.get_obj(key, patch=True)
        source_yaml['automountServiceAccountToken'] = False

    def enrich_deployment_calico_apiserver(self, manifest: Manifest) -> None:
        key = "Deployment_calico-apiserver"
        source_yaml = manifest.get_obj(key, patch=True)
        service_account_name = "calico-apiserver"
        self.enrich_node_selector(manifest, key, plugin_service='apiserver')
        self.enrich_tolerations(manifest, key, plugin_service='apiserver')
        self.enrich_image_for_container(
            manifest, key, plugin_service='apiserver', container_name='calico-apiserver', is_init_container=False)
        self.enrich_resources_for_container(
            manifest, key, plugin_service='apiserver', container_name='calico-apiserver')
        self.enrich_volume_and_volumemount(source_yaml, service_account_name)

        self.enrich_deployment_calico_apiserver_container(manifest)

    def enrich_deployment_calico_apiserver_container(self, manifest: Manifest) -> None:
        key = "Deployment_calico-apiserver"
        # By default, API server searches in the same directory as specified below,
        # but with different file names: apiserver.crt, apiserver.key
        # There is no real necessity to change paths
        # except to make TLS secret creation process the same as for nginx-ingress.
        additional_args = [
            "--tls-cert-file=apiserver.local.config/certificates/tls.crt",
            "--tls-private-key-file=apiserver.local.config/certificates/tls.key"
        ]
        # This also searches in calico.apiserver.args but it is currently not supported by JSON schema.
        self.enrich_args_for_container(
            manifest, key, plugin_service='apiserver', container_name='calico-apiserver',
            extra_args=additional_args)


service_account_secret_calico_apiserver = dedent("""\
    apiVersion: v1
    kind: Secret
    metadata:
      name: calico-apiserver-token
      namespace: calico-apiserver
      annotations:
        kubernetes.io/service-account.name: calico-apiserver
    type: kubernetes.io/service-account-token  
""")



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import io
import itertools
import re
import sys
import threading
import time
from abc import ABC
from copy import deepcopy
from typing import List, Dict, Union, Any, Optional, Mapping, Iterable, IO, Tuple, cast, Callable

import fabric  # type: ignore[import-untyped]
import invoke
import yaml

from kubemarine import system, procedures
from kubemarine.core.cluster import (
    KubernetesCluster, _AnyConnectionTypes, EnrichmentStage, EnrichmentFunction, enrichment
)
from kubemarine.core import connections, static, errors, utils
from kubemarine.core.connections import ConnectionPool
from kubemarine.core.executor import RunnersResult, GenericResult, Token, CommandTimedOut
from kubemarine.core.group import (
    AbstractGroup, NodeGroup, NodeGroupResult, DeferredGroup, RunnersGroupResult,
    GROUP_RUN_TYPE, RemoteExecutor, RunResult
)
from kubemarine.core.resources import DynamicResources

_ShellResult = Dict[str, Any]
_ROLE_SPEC = Union[int, List[str]]


class FakeShell:
    def __init__(self) -> None:
        self.results: Dict[str, List[_ShellResult]] = {}
        self.history: Dict[str, List[_ShellResult]] = {}
        self._lock = threading.Lock()

    def reset(self) -> None:
        self.results = {}
        self.history = {}

    def add(self, results: Mapping[str, GenericResult], do_type: str, args: List[str], usage_limit: int = 0) -> None:
        args.sort()

        for host, result in results.items():
            result = {
                'result': result,
                'do_type': do_type,
                'args': args,
                'used_times': 0
            }

            if usage_limit > 0:
                result['usage_limit'] = usage_limit

            self.results.setdefault(host, []).append(result)

    def find(self, host: str, do_type: str, args: List[str]) -> Optional[GenericResult]:
        # TODO: Support kwargs
        with self._lock:
            results = self.results.get(host, [])
            for i, item in enumerate(results):
                if item['do_type'] == do_type and item['args'] == args:
                    history_found = any(history_item is item for history_item in self.history.get(host, []))
                    if not history_found:
                        self.history.setdefault(host, []).append(item)

                    item['used_times'] += 1
                    if item.get('usage_limit') is not None:
                        item['usage_limit'] -= 1
                        if item['usage_limit'] < 1:
                            del results[i]
                    return item['result']
            return None

    # covered by test.test_demo.TestFakeShell.test_calculate_calls
    def history_find(self, host: str, do_type: str, args: List[str]) -> List[_ShellResult]:
        # TODO: Support kwargs
        result = []
        for item in self.history.get(host, []):
            if item['do_type'] == do_type and item['args'] == args:
                result.append(item)
        return result

    def is_called_each(self, hosts: List[str], do_type: str, args: List[str]) -> bool:
        return all((self.is_called(host, do_type, args) for host in hosts))

    def is_called(self, host: str, do_type: str, args: List[str]) -> bool:
        """
        Returns true if the specified command has already been executed in FakeShell for the specified connection.
        :param host: host to check, for which the desirable command should have been executed.
        :param do_type: The type of required command
        :param args: Required command arguments
        :return: Boolean
        """
        return self.called_times(host, do_type, args) > 0

    def called_times(self, host: str, do_type: str, args: List[str]) -> int:
        """
        Returns number of times the specified command was executed in FakeShell for the specified connection.
        :param host: host to check, for which the desirable command should have been executed.
        :param do_type: The type of required command
        :param args: Required command arguments
        :return: number of calls
        """
        found_entries = self.history_find(host, do_type, args)
        total_used_times: int = sum(found_entry['used_times'] for found_entry in found_entries)
        return total_used_times


class FakeFS:
    def __init__(self) -> None:
        self.storage: Dict[str, Dict[str, str]] = {}
        self.emulate_latency = False
        self._lock = threading.Lock()

    def reset(self) -> None:
        self.storage = {}
        self.emulate_latency = False

    def reset_host(self, host: str) -> None:
        self.storage[host] = {}

    # covered by test.test_demo.TestFakeFS.test_put_file
    # covered by test.test_demo.TestFakeFS.test_put_bytesio
    # covered by test.test_group.TestGroupCall.test_write_stream
    def write(self, host: str, filename: str, data: Union[io.BytesIO, str]) -> None:
        if isinstance(data, io.BytesIO):
            # Emulate how fabric handles file-like objects.
            # See fabric.transfer.Transfer.put
            pointer = data.tell()
            try:
                data.seek(0)
                text = self._transfer(data)
            finally:
                data.seek(pointer)
        else:
            with open(data, "rb") as fl:
                text = self._transfer(fl)

        with self._lock:
            self.storage.setdefault(host, {})[filename] = text

    def _transfer(self, fl: IO) -> str:
        # Emulate how paramiko transfers files.
        # See paramiko.sftp_client.SFTPClient._transfer_with_callback
        target = io.BytesIO()
        while True:
            data = fl.read(32768)
            target.write(data)
            if self.emulate_latency:
                time.sleep(0.1)
            if len(data) == 0:
                break
        return target.getvalue().decode('utf-8')

    # covered by test.test_demo.TestFakeFS.test_put_string
    # covered by test.test_demo.TestFakeFS.test_get_nonexistent
    def read(self, host: str, filename: str) -> Optional[str]:
        return self.storage.get(host, {}).get(filename)

    # covered by test.test_demo.TestFakeFS.test_write_file_to_cluster
    def read_all(self, hosts: List[str], filename: str) -> Dict[str, Optional[str]]:
        result = {}
        for host in hosts:
            result[host] = self.read(host, filename)
        return result


class FakeClusterStorage(utils.ClusterStorage):
    def make_dir(self) -> None:
        pass

    def upload_and_rotate(self) -> None:
        cluster = cast(FakeKubernetesCluster, self.cluster)
        cluster.uploaded_archives.append(self.local_archive_path)

    def upload_info_new_control_planes(self) -> None:
        pass


class FakeKubernetesCluster(KubernetesCluster):

    def __init__(self, *args: Any, **kwargs: Any):
        self.resources: FakeClusterResources = kwargs.pop("resources")
        self.fake_shell = self.resources.fake_shell
        self.fake_fs = self.resources.fake_fs
        self.uploaded_archives: List[str] = []
        super().__init__(*args, **kwargs)

    def _create_connection_pool(self, nodes: Dict[str, dict], gateway_nodes: Dict[str, dict], hosts: List[str]) -> ConnectionPool:
        return FakeConnectionPool(nodes, gateway_nodes, hosts, self.fake_shell, self.fake_fs)

    def _create_cluster_storage(self, context: dict) -> utils.ClusterStorage:
        return FakeClusterStorage(self, context)

    def make_group(self, ips: Iterable[_AnyConnectionTypes]) -> FakeNodeGroup:
        return FakeNodeGroup(ips, self)


class FakeClusterResources(DynamicResources):
    def __init__(self, context: dict,
                 *,
                 nodes_context: Dict[str, Any] = None,
                 fake_shell: FakeShell = None, fake_fs: FakeFS = None):
        super().__init__(context)
        self.fake_shell = fake_shell if fake_shell else FakeShell()
        self.fake_fs = fake_fs if fake_fs else FakeFS()

        self.finalized_inventory: dict = {}

        self._enrichment_functions = super().enrichment_functions()
        # Let's do not assign self._nodes_context directly to make it more close to the real enrichment.
        if nodes_context is not None:
            fn_idx = self._enrichment_functions.index(system.detect_nodes_context)
            self._enrichment_functions[fn_idx] = enrichment(EnrichmentStage.LIGHT)\
                (lambda c: c.nodes_context.update(nodes_context))

    @property
    def working_inventory(self) -> dict:
        cluster = self.cluster_if_initialized()
        if cluster is not None:
            return cluster.inventory

        return {}

    def _store_finalized_inventory(self, finalized_inventory: dict, finalized_filename: str) -> None:
        self.finalized_inventory = finalized_inventory
        super()._store_finalized_inventory(finalized_inventory, finalized_filename)

    def cluster_if_initialized(self) -> Optional[FakeKubernetesCluster]:
        cluster = super().cluster_if_initialized()
        return None if cluster is None else cast(FakeKubernetesCluster, cluster)

    def cluster(self, stage: EnrichmentStage = EnrichmentStage.PROCEDURE) -> FakeKubernetesCluster:
        return cast(FakeKubernetesCluster, super().cluster(stage))

    def _new_cluster_instance(self, context: dict) -> FakeKubernetesCluster:
        return FakeKubernetesCluster(
            self.inventory(), context, self.procedure_inventory(),
            self.logger(),
            connection_pool=self._connection_pool, nodes_context=self._nodes_context,
            resources=self,
        )

    def insert_enrichment_function(self, near: EnrichmentFunction, stages: EnrichmentStage,
                                   fn: Callable[[KubernetesCluster], Optional[dict]],
                                   *,
                                   procedure: str = None,
                                   after: bool = False) -> None:

        enrichment_fn = enrichment(stages, procedures=None if procedure is None else [procedure])(fn)

        idx = self._enrichment_functions.index(near)
        if after:
            idx += 1
        self._enrichment_functions.insert(idx, enrichment_fn)

        self._skip_default_enrichment = None

    def enrichment_functions(self) -> List[EnrichmentFunction]:
        return self._enrichment_functions


class FakeResources(FakeClusterResources):
    def __init__(self, context: dict, inventory: dict = None,
                 *,
                 procedure_inventory: dict = None,
                 nodes_context: Dict[str, Any] = None,
                 fake_shell: FakeShell = None, fake_fs: FakeFS = None):
        super().__init__(context, nodes_context=nodes_context, fake_shell=fake_shell, fake_fs=fake_fs)
        self.inventory_filepath = None
        self.procedure_inventory_filepath = None
        self._inventory = inventory
        self._procedure_inventory = procedure_inventory

    def _store_inventory(self, inventory: dict) -> None:
        pass

    def _store_finalized_inventory(self, finalized_inventory: dict, finalized_filename: str) -> None:
        self.finalized_inventory = finalized_inventory
        utils.dump_file(self, yaml.dump(finalized_inventory), finalized_filename)


class FakeConnection(connections.Connection):

    def __init__(self, ip: str, fake_shell: FakeShell, fake_fs: FakeFS, **kw: Any):
        super().__init__(ip, **kw)
        self.fake_shell = fake_shell
        self.fake_fs = fake_fs

        command_sep = r'[=\-_]{32}'
        sep_symbol = r'\&\&|;'
        final_sep = rf" ({sep_symbol}) " \
                    rf"printf \"%s\\n\$\?\\n%s\\n\" \"({command_sep})\" \"\2\" \1 " \
                    rf"echo \"\2\" 1>\&2 \1 (sudo )?"

        self.separator_ptrn = re.compile(final_sep)

    def __setattr__(self, key: str, value: Any) -> None:
        # fabric Connection has special handling of this method. Call default behaviour for custom attributes.
        if key in ('fake_shell', 'fake_fs', 'separator_ptrn'):
            return object.__setattr__(self, key, value)
        super().__setattr__(key, value)

    def run(self, command: str, **kwargs: Any) -> fabric.runners.Result:
        return self._do("run", command, **kwargs)

    def sudo(self, command: str, **kwargs: Any) -> fabric.runners.Result:
        return self._do("sudo", command, **kwargs)

    # not implemented
    def get(self, remote_file: str, local_file: str, **kwargs: Any) -> None:  # pylint: disable=arguments-differ
        raise NotImplementedError("Stub for fabric Connection.get() is not implemented")

    def put(self, data: Union[io.BytesIO, str], filename: str, **kwargs: Any) -> None:  # pylint: disable=arguments-differ
        # It should return fabric.transfer.Result, but currently returns None.
        # Transfer Result is currently never handled.
        self.fake_fs.write(self.host, filename, data)

    def _do(self, do_type: str, original_command: str, **kwargs: Any) -> fabric.runners.Result:
        # start fake execution of commands
        commands, sep_symbol, command_sep = self._split_command(do_type, original_command)

        final_res = fabric.runners.Result(stdout="", stderr="", exited=None,
                                          connection=self, command=original_command)
        prev_exited = None
        i = 0
        for command in commands:
            found_result = self.fake_shell.find(self.host, do_type, [command])

            if found_result is None:
                raise Exception('Fake result not found for requested action type \'%s\' and command %s' % (do_type, [command]))

            timeout_exception = None
            if isinstance(found_result, BaseException):
                if i > 0:
                    raise ValueError("Exception can be thrown only for the whole command")
                elif isinstance(found_result, CommandTimedOut):
                    timeout_exception = found_result
                    found_result = found_result.result
                else:
                    raise found_result

            hide = kwargs.get('hide', False)
            if found_result.hide != hide:
                raise Exception(f"Fake result has hide={found_result.hide} while hide={hide} was requested")

            if i > 0:
                final_res.stdout += command_sep + '\n' + str(prev_exited) + '\n' + command_sep + '\n'
                final_res.stderr += command_sep + '\n'
            i += 1

            for stream_t in ('stdout', 'stderr'):
                output = getattr(found_result, stream_t)
                if output and not hide:
                    stream = getattr(sys, stream_t)
                    stream.write(output)
                    stream.flush()

                final_output = getattr(final_res, stream_t)
                setattr(final_res, stream_t, final_output + output)

            final_res.exited = prev_exited = found_result.exited

            if timeout_exception is not None:
                raise invoke.CommandTimedOut(final_res, timeout_exception.timeout)

            if prev_exited != 0 and sep_symbol != ';':
                # stop fake execution
                break

        if prev_exited == 0 or kwargs.get('warn', False):
            return final_res

        raise invoke.UnexpectedExit(final_res)

    def _split_command(self, do_type: str, command: str) -> Tuple[List[str], str, str]:
        """
        This is a reverse operation to the RemoteExecutor#_merge_actions
        """
        tokens = self.separator_ptrn.split(command)
        commands = []
        i = 0
        sep_symbol = None
        command_sep = None
        while i < len(tokens):
            commands.append(tokens[i])
            i += 1
            if i < len(tokens):
                sep_symbol = self._compare_and_return(sep_symbol, tokens[i], "Separator symbols are not equal")
                command_sep = self._compare_and_return(command_sep, tokens[i + 1], "Command separators are not equal")
                resolved_do_type = "sudo" if tokens[i + 2] == 'sudo ' else "run"
                self._compare_and_return(do_type, resolved_do_type, "Do types are not equal")
                i += 3

        return commands, sep_symbol or '', command_sep or ''

    def _compare_and_return(self, one: Optional[str], another: str, msg: str) -> str:
        if one is None or one == another:
            return another
        else:
            raise ValueError(msg)

    def close(self) -> None:
        pass


class FakeAbstractGroup(AbstractGroup[GROUP_RUN_TYPE], ABC):
    def _put_with_mv(self, local_stream: Union[bytes, str], remote_file: str,
                     backup: bool, sudo: bool, mkdir: bool, immutable: bool) -> None:
        super()._put_with_mv(local_stream, remote_file, backup=False, sudo=False, mkdir=False, immutable=False)


class FakeNodeGroup(NodeGroup, FakeAbstractGroup[RunnersGroupResult]):
    def _make_group(self, ips: Iterable[Union[str, NodeGroup]]) -> FakeNodeGroup:
        return FakeNodeGroup(ips, self.cluster)

    def _make_defer(self, executor: RemoteExecutor) -> FakeDeferredGroup:
        return FakeDeferredGroup(self.nodes, self.cluster, executor)

    def get_local_file_sha1(self, local_file: Union[bytes, str]) -> str:
        return '0'

    def get_remote_file_sha1(self, filename: str) -> Dict[str, Optional[str]]:
        return {host: '1' for host in self.nodes}


class FakeDeferredGroup(DeferredGroup, FakeAbstractGroup[Token]):
    def _make_group(self, ips: Iterable[Union[str, DeferredGroup]]) -> FakeDeferredGroup:
        return FakeDeferredGroup(ips, self.cluster, self._executor)


class FakeConnectionPool(connections.ConnectionPool):
    def __init__(self, nodes: Dict[str, dict], gateway_nodes: Dict[str, dict], hosts: List[str],
                 fake_shell: FakeShell, fake_fs: FakeFS):
        self.fake_shell = fake_shell
        self.fake_fs = fake_fs
        super().__init__(nodes, gateway_nodes, hosts)

    def _create_connection_from_details(self, ip: str, conn_details: dict,
                                        gateway: fabric.connection.Connection = None,
                                        inline_ssh_env: bool = True) -> FakeConnection:
        return FakeConnection(
            ip, self.fake_shell, self.fake_fs, gateway=gateway,
            user=conn_details.get('username', static.GLOBALS['connection']['defaults']['username']),
            port=conn_details.get('connection_port', static.GLOBALS['connection']['defaults']['port'])
        )


def create_silent_context(args: list = None, procedure: str = 'install') -> dict:
    args = list(args) if args else []
    # todo probably increase logging level to get rid of spam in logs.

    context: dict = procedures.import_procedure(procedure).create_context(args)
    context['preserve_inventory'] = False
    context['make_finalized_inventory'] = False
    context['load_inventory_silent'] = True

    parsed_args: dict = context['execution_arguments']
    parsed_args['disable_dump'] = True
    del parsed_args['ansible_inventory_location']

    return context


def new_resources(inventory: dict, procedure_inventory: dict = None, context: dict = None,
                  nodes_context: dict = None) -> FakeResources:
    if context is None:
        context = create_silent_context()

    nds_context = generate_nodes_context(inventory, procedure_inventory, context)
    if nodes_context is not None:
        nds_context.update(nodes_context)

    return FakeResources(context, inventory,
                         procedure_inventory=procedure_inventory, nodes_context=nds_context)


def new_cluster(inventory: dict, procedure_inventory: dict = None, context: dict = None,
                nodes_context: dict = None) -> FakeKubernetesCluster:
    res = new_resources(inventory, procedure_inventory, context, nodes_context)
    try:
        return res.cluster()
    except errors.FailException as exc:
        if exc.reason is not None:
            raise exc.reason

        raise


def generate_node_context(*,
                          online: bool = True, accessible: bool = True, sudo: str = 'Root',
                          os_name: str = 'centos', os_version: str = '7.9',
                          net_interface: str = 'eth0') -> dict:
    os_family = system.detect_os_family_by_name_version(os_name, os_version)

    if not online:
        accessible = False
    if not accessible:
        sudo = 'No'
        os_name, os_version, os_family, net_interface = tuple('<undefined>' for _ in range(4))

    return {
        'access': {
            'online': online,
            'accessible': accessible,
            'sudo': sudo
        },
        'active_interface': net_interface,
        'os': {
            'name': os_name,
            'family': os_family,
            'version': os_version
        }
    }


def generate_nodes_context(inventory: dict, procedure_inventory: dict = None, context: dict = None,
                           *,
                           os_name: str = 'centos', os_version: str = '7.9',
                           net_interface: str = 'eth0') -> dict:
    procedure = 'install' if context is None else context['initial_procedure']
    if procedure_inventory is None:
        procedure_inventory = {}

    nodes = inventory['nodes']
    if procedure == 'add_node':
        nodes = itertools.chain(nodes, procedure_inventory.get('nodes', []))

    context = {}
    for node in nodes:
        connect_to = node['internal_address']
        if node.get('address'):
            connect_to = node['address']
        context[connect_to] = generate_node_context(os_name=os_name, os_version=os_version, net_interface=net_interface)

    return context


def generate_inventory(balancer: _ROLE_SPEC = 1, control_plane: _ROLE_SPEC = 1, worker: _ROLE_SPEC = 1,
                       keepalived: _ROLE_SPEC = 0, haproxy_mntc: _ROLE_SPEC = 0) -> dict:
    inventory: dict = {
        'node_defaults': {
            'keyfile': '/dev/null',
            'username': 'anonymous'
        },
        'nodes': [],
        'services': {
            'cri': {}
        },
        'cluster_name': 'k8s.fake.local'
    }

    id_roles_map: Dict[str, List[str]] = {}

    for role_name, item in (('balancer', balancer), ('control-plane', control_plane), ('worker', worker)):

        if isinstance(item, int):
            item = [f'{role_name}-{i + 1}' for i in range(item)]

        for id_ in item:
            id_roles_map.setdefault(id_, []).append(role_name)

    ip_i = 0

    for id_, roles in id_roles_map.items():
        ip_i = ip_i + 1
        if "control-plane" in roles and worker == 0:
            roles.append('worker')
        inventory['nodes'].append({
            'name': id_,
            'address': '10.101.1.%s' % ip_i,
            'internal_address': '192.168.0.%s' % ip_i,
            'roles': roles
        })

    ip_i = 0
    vrrp_ips: List[Union[str, dict]] = []

    if isinstance(keepalived, list):
        vrrp_ips.extend(deepcopy(keepalived))
    elif isinstance(keepalived, int) and keepalived > 0:
        for _ in range(keepalived):
            ip_i = ip_i + 1
            vrrp_ips.append('10.101.2.%s' % ip_i)

    if isinstance(haproxy_mntc, list):
        vrrp_ips.extend(deepcopy(haproxy_mntc))
    elif isinstance(haproxy_mntc, int) and haproxy_mntc > 0:
        for _ in range(haproxy_mntc):
            ip_i = ip_i + 1
            vrrp_ips.append({
                'ip': '10.101.2.%s' % ip_i,
                'params': {
                    'maintenance-type': 'not bind'
                }
            })

    if haproxy_mntc:
        inventory['services'] = {
            'loadbalancer' : {
                'haproxy': {
                    'maintenance_mode': True
                }
            }
        }

    inventory['vrrp_ips'] = vrrp_ips

    return inventory


def generate_procedure_inventory(procedure: str) -> dict:
    procedure_inventory: dict = {}
    # set some commonly required properties
    if procedure == 'manage_pss':
        procedure_inventory['pss'] = {'pod-security': 'enabled'}

    return procedure_inventory


def create_nodegroup_result_by_hosts(cluster: KubernetesCluster, results: Dict[str, GenericResult]) -> NodeGroupResult:
    return NodeGroupResult(cluster, results)


def create_exception_result(group_: AbstractGroup[RunResult], exception: Exception) -> NodeGroupResult:
    hosts_to_result = create_hosts_exception_result(group_.get_hosts(), exception)
    return create_nodegroup_result_by_hosts(group_.cluster, hosts_to_result)


def create_hosts_exception_result(hosts: List[str], exception: Exception) -> Dict[str, GenericResult]:
    return {host: exception for host in hosts}


def create_nodegroup_result(group_: AbstractGroup[RunResult], stdout: str = '', stderr: str = '', hide: bool = True,
                            code: int = 0, timeout: int = None) -> NodeGroupResult:
    hosts_to_result = create_hosts_result(group_.get_hosts(), stdout, stderr, hide, code, timeout)
    return create_nodegroup_result_by_hosts(group_.cluster, hosts_to_result)


def create_hosts_result(hosts: List[str], stdout: str = '', stderr: str = '', hide: bool = True,
                        code: int = 0, timeout: int = None) -> Dict[str, GenericResult]:
    # each host should have its own result instance.
    return {host: create_result(stdout, stderr, hide, code, timeout) for host in hosts}


def create_result(stdout: str = '', stderr: str = '', hide: bool = True,
                  code: int = 0, timeout: int = None) -> GenericResult:
    # command and 'hide' option will be later replaced with actual
    result = RunnersResult(["fake command"], [code], stdout=stdout, stderr=stderr, hide=hide)
    if timeout is None:
        return result

    return CommandTimedOut(result, timeout)


def new_scheme(scheme: dict, role: str, number: int) -> dict:
    scheme = deepcopy(scheme)
    scheme[role] = number
    return scheme


FULLHA: Dict[str, _ROLE_SPEC] = {'balancer': 1, 'control_plane': 3, 'worker': 3}
FULLHA_KEEPALIVED: Dict[str, _ROLE_SPEC] = {'balancer': 2, 'control_plane': 3, 'worker': 3, 'keepalived': 1}
FULLHA_NOBALANCERS: Dict[str, _ROLE_SPEC] = {'balancer': 0, 'control_plane': 3, 'worker': 3}
ALLINONE: Dict[str, _ROLE_SPEC] = {
    'control_plane': 1, 'balancer': ['control-plane-1'], 'worker': ['control-plane-1'],
    'keepalived': 1}
MINIHA: Dict[str, _ROLE_SPEC] = {'control_plane': 3}
MINIHA_KEEPALIVED: Dict[str, _ROLE_SPEC] = {
    'control_plane': 3,
    'balancer': ['control-plane-1', 'control-plane-2', 'control-plane-3'],
    'worker': ['control-plane-1', 'control-plane-2', 'control-plane-3'],
    'keepalived': 1}
NON_HA_BALANCER: Dict[str, _ROLE_SPEC] = {
    'balancer': 1, 'control_plane': 3, 'worker': ['control-plane-1', 'control-plane-2', 'control-plane-3']}



#!/usr/bin/env python3
# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import OrderedDict
from types import FunctionType
from typing import Callable, List, Dict, cast

from kubemarine.core.cluster import KubernetesCluster
from kubemarine.core.errors import KME
from kubemarine import (
    system, sysctl, haproxy, keepalived, kubernetes, plugins,
    kubernetes_accounts, selinux, thirdparties, audit, coredns, cri, packages, apparmor, modprobe
)
from kubemarine.core import flow, utils, summary
from kubemarine.core.group import NodeGroup, RunnersGroupResult, CollectorCallback
from kubemarine.core.resources import DynamicResources


TASK_CALLABLE = Callable[[KubernetesCluster], None]
DECORATED_GROUP_CALLABLE = Callable[[NodeGroup], None]


def _applicable_for_new_nodes_with_roles(*roles: str) -> Callable[[DECORATED_GROUP_CALLABLE], TASK_CALLABLE]:
    """
    Decorator to annotate installation methods.
    If there are no new nodes with the specified roles to be added / installed to the cluster,
    the decorator skips execution of the method.
    Otherwise, it runs the annotated method with the calculated group of nodes with the specified roles.
    Note that the signature of annotated method should be f(NodeGroup),
    but the resulting wrapping method will be f(KubernetesCluster).

    :param roles: roles of nodes for which the annotated method is applicable.
    :return: new wrapping method.
    """
    if not roles:
        raise Exception(f'Roles are not defined')

    def roles_wrapper(fn: DECORATED_GROUP_CALLABLE) -> TASK_CALLABLE:
        def cluster_wrapper(cluster: KubernetesCluster) -> None:
            candidate_group = cluster.get_new_nodes_or_self()
            group = cluster.make_group_from_roles(roles)
            group = group.intersection_group(candidate_group)
            if not group.is_empty():
                fn(group)
            else:
                func = cast(FunctionType, fn)
                fn_name = func.__module__ + '.' + func.__qualname__
                cluster.log.debug(f"Skip running {fn_name} as no new node with roles {roles} has been found.")

        return cluster_wrapper

    return roles_wrapper


def system_prepare_check_sudoer(cluster: KubernetesCluster) -> None:
    group = cluster.make_group_from_roles(['control-plane', 'balancer']).include_group(cluster.get_new_nodes_or_self())
    not_sudoers = []
    for host, node_context in cluster.nodes_context.items():
        if host not in group.nodes:
            continue
        access_info = node_context['access']
        if access_info['online'] and access_info['sudo'] == 'Root':
            cluster.log.debug("%s online and has root" % host)
        else:
            not_sudoers.append(host)

    if not_sudoers:
        raise KME("KME0005", hostnames=not_sudoers)


@_applicable_for_new_nodes_with_roles('all')
def system_prepare_check_system(group: NodeGroup) -> None:
    cluster: KubernetesCluster = group.cluster
    cluster.log.debug(system.fetch_os_versions(cluster))
    for address, context in cluster.nodes_context.items():
        if address not in group.nodes:
            continue
        if context["os"]["family"] == "unsupported":
            raise Exception('%s host operating system is unsupported' % address)
        if context["os"]["family"] == "unknown":
            supported_os_versions = []
            os_family_list = cluster.globals["compatibility_map"]["distributives"][context["os"]["name"]]
            for os_family_item in os_family_list:
                supported_os_versions.extend(os_family_item["versions"])
            raise Exception("%s running on unknown %s version. "
                            "Expected %s, got '%s'" % (address,
                                                       context["os"]["name"],
                                                       supported_os_versions,
                                                       context["os"]["version"]))


def system_prepare_check_cluster_installation(cluster: KubernetesCluster) -> None:
    if kubernetes.is_cluster_installed(cluster):
        cluster.log.debug('Cluster already installed and available at %s' % cluster.context['controlplain_uri'])
    else:
        cluster.log.debug('There is no any installed cluster')


@_applicable_for_new_nodes_with_roles('all')
def system_prepare_system_chrony(group: NodeGroup) -> None:
    cluster: KubernetesCluster = group.cluster
    if cluster.inventory['services']['ntp'].get('chrony', {}).get('servers') is None:
        cluster.log.debug("Skipped - NTP servers from chrony is not defined in config file")
        return
    group.call(system.configure_chronyd)


@_applicable_for_new_nodes_with_roles('all')
def system_prepare_system_timesyncd(group: NodeGroup) -> None:
    cluster: KubernetesCluster = group.cluster
    if not cluster.inventory['services']['ntp'].get('timesyncd', {}).get('Time', {}).get('NTP') and \
            not cluster.inventory['services']['ntp'].get('timesyncd', {}).get('Time', {}).get('FallbackNTP'):
        cluster.log.debug("Skipped - NTP servers from timesyncd is not defined in config file")
        return
    group.call(system.configure_timesyncd)


@_applicable_for_new_nodes_with_roles('all')
def system_prepare_system_sysctl(group: NodeGroup) -> None:
    is_updated = system.configure_sensitive_service(group, sysctl.setup_sysctl)
    if is_updated:
        group.call(system.verify_sysctl)


@_applicable_for_new_nodes_with_roles('all')
def system_prepare_system_setup_selinux(group: NodeGroup) -> None:
    system.configure_sensitive_service(group, selinux.setup_selinux)


@_applicable_for_new_nodes_with_roles('all')
def system_prepare_system_setup_apparmor(group: NodeGroup) -> None:
    system.configure_sensitive_service(group, apparmor.setup_apparmor)


@_applicable_for_new_nodes_with_roles('all')
def system_prepare_system_disable_firewalld(group: NodeGroup) -> None:
    group.call(system.disable_firewalld)


@_applicable_for_new_nodes_with_roles('all')
def system_prepare_system_disable_swap(group: NodeGroup) -> None:
    group.call(system.disable_swap)


@_applicable_for_new_nodes_with_roles('all')
def system_prepare_system_modprobe(group: NodeGroup) -> None:
    system.configure_sensitive_service(group, modprobe.setup_modprobe)


@_applicable_for_new_nodes_with_roles('control-plane', 'worker')
def system_install_audit(group: NodeGroup) -> None:
    group.call(audit.install)


@_applicable_for_new_nodes_with_roles('control-plane', 'worker')
def system_prepare_audit(group: NodeGroup) -> None:
    group.call(audit.apply_audit_rules)


@_applicable_for_new_nodes_with_roles('control-plane')
def deploy_kubernetes_audit(group: NodeGroup) -> None:
    """
    Task generates rules for logging kubernetes and on audit
    """
    cluster = group.cluster
    # kubernetes api-server is already installed with target audit configuration for install/add_node procedures.
    if cluster.is_task_completed('deploy.kubernetes.init'):
        cluster.log.debug("Kubernetes audit policy is already configured")
        return

    kubernetes.prepare_audit_policy(group)
    group.call(kubernetes.components.reconfigure_components, components=['kube-apiserver'], force_restart=True)


@_applicable_for_new_nodes_with_roles('all')
def system_prepare_dns_hostname(group: NodeGroup) -> None:
    cluster: KubernetesCluster = group.cluster
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            cluster.log.debug("Changing hostname '%s' = '%s'" % (node.get_host(), node.get_node_name()))
            node.sudo("hostnamectl set-hostname %s" % node.get_node_name())


@_applicable_for_new_nodes_with_roles('all')
def system_prepare_dns_resolv_conf(group: NodeGroup) -> None:
    cluster: KubernetesCluster = group.cluster
    if cluster.inventory["services"].get("resolv.conf") is None:
        cluster.log.debug("Skipped - resolv.conf section not defined in config file")
        return

    system.update_resolv_conf(group, config=cluster.inventory["services"].get("resolv.conf"))
    cluster.log.debug(group.sudo("ls -la /etc/resolv.conf; sudo lsattr /etc/resolv.conf"))


def system_prepare_dns_etc_hosts(cluster: KubernetesCluster) -> None:
    remained_offline = cluster.nodes['all'].get_online_nodes(False)
    if not remained_offline.is_empty():
        if not cluster.inventory['globals']['ignore_unavailable_nodes_for_etchosts_update']:
            # generally all the nodes must be available during this task running
            raise Exception("Nodes %s are not reachable" % remained_offline.get_hosts())
        else:
            # in some cases we allow to update /etc/hosts only at the available nodes
            cluster.log.warning("Nodes %s are not reachable. You can update their /etc/hosts by running install job "
                "with the task prepare.dns.etc_hosts only when these nodes become available." % remained_offline.get_hosts())

    config = system.generate_etc_hosts_config(cluster.inventory, 'etc_hosts')
    config += system.generate_etc_hosts_config(cluster.inventory, 'etc_hosts_generated')

    utils.dump_file(cluster, config, 'etc_hosts')
    cluster.log.debug("\nUploading...")

    if not cluster.inventory['globals']['ignore_unavailable_nodes_for_etchosts_update']:
        # generally we update /etc/hosts at all the nodes
        group = cluster.nodes['all']
    else:
        # update /etc/hosts only at the available nodes
        group = cluster.nodes['all'].get_online_nodes(True)

    system.update_etc_hosts(group, config=config)
    cluster.log.debug(group.sudo("ls -la /etc/hosts"))


@_applicable_for_new_nodes_with_roles('all')
def system_prepare_package_manager_configure(group: NodeGroup) -> None:
    cluster: KubernetesCluster = group.cluster
    repositories = cluster.inventory['services']['packages']['package_manager'].get("repositories")
    if not repositories:
        cluster.log.debug("Skipped - no repositories defined for configuration")
        return

    group.call(packages.backup_repo)
    group.call(packages.add_repo, repo_data=repositories)

    cluster.log.debug("Nodes contain the following repositories:")
    cluster.log.debug(packages.ls_repofiles(group))


@_applicable_for_new_nodes_with_roles('all')
def system_prepare_package_manager_disable_unattended_upgrades(group: NodeGroup) -> None:
    group.call(packages.disable_unattended_upgrade)


@_applicable_for_new_nodes_with_roles('all')
def system_prepare_package_manager_manage_packages(group: NodeGroup) -> None:
    group.call_batch([
        manage_mandatory_packages,
        manage_custom_packages
    ])


def manage_mandatory_packages(group: NodeGroup) -> RunnersGroupResult:
    cluster: KubernetesCluster = group.cluster

    collector = CollectorCallback(cluster)
    with group.new_executor() as exe:
        for node in exe.group.get_ordered_members_list():
            pkgs: List[str] = []
            for package in cluster.inventory["services"]["packages"]['mandatory'].keys():
                hosts_to_packages = packages.get_association_hosts_to_packages(node, cluster.inventory, package)
                pkgs.extend(next(iter(hosts_to_packages.values()), []))

            if pkgs:
                cluster.log.debug(f"Installing {pkgs} on {node.get_node_name()!r}")
                packages.install(node, include=pkgs, pty=True, callback=collector)

    return collector.result


def manage_custom_packages(group: NodeGroup) -> None:
    cluster: KubernetesCluster = group.cluster

    batch_results: Dict[str, RunnersGroupResult] = {}
    packages_section = cluster.inventory["services"].get("packages", {})
    if packages_section.get("remove", {}).get("include"):
        cluster.log.debug("Running kubemarine.packages.remove: ")
        remove = packages_section['remove']
        batch_results['remove'] = results = packages.remove(
            group, include=remove['include'], exclude=remove.get('exclude'), pty=True)
        cluster.log.debug(results)

    if packages_section.get("install", {}).get("include"):
        cluster.log.debug("Running kubemarine.packages.install: ")
        install = packages_section['install']
        batch_results['install'] = results = packages.install(
            group, include=install['include'], exclude=install.get('exclude'), pty=True)
        cluster.log.debug(results)

    if packages_section.get("upgrade", {}).get("include"):
        cluster.log.debug("Running kubemarine.packages.upgrade: ")
        upgrade = packages_section['upgrade']
        batch_results['upgrade'] = results = packages.upgrade(
            group, include=upgrade['include'], exclude=upgrade.get('exclude'), pty=True)
        cluster.log.debug(results)

    if not batch_results:
        cluster.log.debug("Skipped - no packages configuration defined in config file")
        return None

    any_changes_found = False
    for action, results in batch_results.items():
        cluster.log.verbose('Verifying packages changes after \'%s\' action...' % action)
        for host, result in results.items():
            node = cluster.make_group([host])
            if not packages.no_changes_found(node, action, result):
                cluster.log.verbose('Packages changed at %s' % host)
                any_changes_found = True

    if any_changes_found:
        cluster.log.verbose('Packages changed, scheduling nodes restart...')
        cluster.schedule_cumulative_point(system.reboot_nodes)
    else:
        cluster.log.verbose('No packages changed, nodes restart will not be scheduled')

    return None


@_applicable_for_new_nodes_with_roles('control-plane', 'worker')
def system_cri_install(group: NodeGroup) -> None:
    """
    Task which is used to install CRI. Could be skipped, if CRI already installed.
    """
    group.call(cri.install)


@_applicable_for_new_nodes_with_roles('control-plane', 'worker')
def system_cri_configure(group: NodeGroup) -> None:
    """
    Task which is used to configure CRI. Could be skipped, if CRI already configured.
    """
    group.call(cri.configure)


@_applicable_for_new_nodes_with_roles('all')
def system_prepare_thirdparties(group: NodeGroup) -> None:
    cluster: KubernetesCluster = group.cluster
    if not cluster.inventory['services'].get('thirdparties', {}):
        cluster.log.debug("Skipped - no thirdparties defined in config file")
        return

    group.call(thirdparties.install_all_thirparties)


@_applicable_for_new_nodes_with_roles('balancer')
def deploy_loadbalancer_haproxy_install(group: NodeGroup) -> None:
    group.call(haproxy.install)


def get_haproxy_configure_group(cluster: KubernetesCluster) -> NodeGroup:
    changed_nodes = cluster.get_changed_nodes()
    balancers = cluster.make_group_from_roles(['balancer'])
    if not changed_nodes.having_roles(['control-plane', 'worker']).is_empty():
        return balancers
    elif cluster.context['initial_procedure'] != 'remove_node':
        new_nodes = cluster.get_new_nodes_or_self()
        return balancers.intersection_group(new_nodes)
    else:
        return cluster.make_group([])


def deploy_loadbalancer_haproxy_configure(cluster: KubernetesCluster) -> None:

    if not cluster.inventory['services'].get('loadbalancer', {}) \
            .get('haproxy', {}).get('keep_configs_updated', True):
        cluster.log.debug('Skipped - haproxy balancers configs update manually disabled')
        return

    group = get_haproxy_configure_group(cluster)

    if group.is_empty():
        cluster.log.debug('Skipped - no balancers to perform')
        return

    with group.new_executor() as exe:
        exe.group.call_batch([
            haproxy.configure,
            haproxy.override_haproxy18,
        ])

    haproxy.restart(group)


@_applicable_for_new_nodes_with_roles('keepalived')
def deploy_loadbalancer_keepalived_install(group: NodeGroup) -> None:
    group.call(keepalived.install)


def get_keepalived_configure_group(cluster: KubernetesCluster) -> NodeGroup:
    # For install procedure, configure all keepalives.
    # If balancer with VRPP IP is added or removed, reconfigure all keepalives
    changed_nodes = cluster.get_changed_nodes()
    if cluster.context['initial_procedure'] != 'install' and changed_nodes.having_roles(['keepalived']).is_empty():
        return cluster.make_group([])
    else:
        return cluster.make_group_from_roles(['keepalived'])


def deploy_loadbalancer_keepalived_configure(cluster: KubernetesCluster) -> None:
    if not cluster.inventory['services'].get('loadbalancer', {}) \
            .get('keepalived', {}).get('keep_configs_updated', True):
        cluster.log.debug('Skipped - keepalived balancers configs update manually disabled')
        return

    group = get_keepalived_configure_group(cluster)

    if group.is_empty():
        cluster.log.debug('Skipped - no VRRP IPs to perform')
        return

    group.call(keepalived.configure)


@_applicable_for_new_nodes_with_roles('control-plane', 'worker')
def deploy_kubernetes_reset(group: NodeGroup) -> None:
    group.call(kubernetes.reset_installation_env)


@_applicable_for_new_nodes_with_roles('control-plane', 'worker')
def deploy_kubernetes_install(group: NodeGroup) -> None:
    group.cluster.log.debug("Setting up Kubernetes...")
    group.call(kubernetes.install)


@_applicable_for_new_nodes_with_roles('control-plane', 'worker')
def deploy_kubernetes_prepull_images(group: NodeGroup) -> None:
    group.cluster.log.debug("Prepulling Kubernetes images...")
    group.call(kubernetes.images_grouped_prepull)


def deploy_kubernetes_init(cluster: KubernetesCluster) -> None:
    cluster.nodes['control-plane'].call_batch([
        kubernetes.init_first_control_plane,
        kubernetes.join_other_control_planes
    ])

    if 'worker' in cluster.nodes:
        cluster.nodes['worker'].exclude_group(cluster.nodes['control-plane']) \
            .call(kubernetes.init_workers)

    cluster.nodes['all'].call_batch([
        kubernetes.apply_labels,
        kubernetes.apply_taints
    ])

    kubernetes.schedule_running_nodes_report(cluster)


def deploy_coredns(cluster: KubernetesCluster) -> None:
    config = coredns.generate_configmap(cluster.inventory)

    cluster.log.debug('Applying patch...')
    cluster.log.debug(coredns.apply_patch(cluster))

    cluster.log.debug('Applying configmap...')
    cluster.log.debug(coredns.apply_configmap(cluster, config))


def deploy_plugins(cluster: KubernetesCluster) -> None:
    plugins.install(cluster)


def deploy_accounts(cluster: KubernetesCluster) -> None:
    kubernetes_accounts.install(cluster)


def overview(cluster: KubernetesCluster) -> None:
    cluster.log.debug("Retrieving cluster status...")
    control_plane = cluster.nodes["control-plane"].get_first_member()
    cluster.log.debug("\nNAMESPACES:")
    control_plane.sudo("kubectl get namespaces", hide=False)
    cluster.log.debug("\nNODES:")
    control_plane.sudo("kubectl get nodes -o wide", hide=False)
    cluster.log.debug("\nPODS:")
    control_plane.sudo("kubectl get pods -A -o=wide", hide=False)
    cluster.log.debug("\nREPLICA SETS:")
    control_plane.sudo("kubectl get rs -A", hide=False)
    cluster.log.debug("\nDAEMON SETS:")
    control_plane.sudo("kubectl get ds -A", hide=False)
    cluster.log.debug("\nSERVICES:")
    control_plane.sudo("kubectl get svc -A -o wide", hide=False)
    cluster.log.debug("\nINGRESS:")
    control_plane.sudo("kubectl get ing -A -o wide", hide=False)
    cluster.log.debug("\nDESCRIPTION:")
    control_plane.sudo("kubectl describe nodes", hide=False)
    cluster.log.debug("\n")
    control_plane.sudo("kubectl cluster-info", hide=False, warn=True)


tasks = OrderedDict({
    "prepare": {
        "check": {
            "sudoer": system_prepare_check_sudoer,
            "system": system_prepare_check_system,
            "cluster_installation": system_prepare_check_cluster_installation
        },
        "dns": {
            "hostname": system_prepare_dns_hostname,
            "etc_hosts": system_prepare_dns_etc_hosts,
            "resolv_conf": system_prepare_dns_resolv_conf
        },
        "package_manager": {
            "configure": system_prepare_package_manager_configure,
            "disable_unattended_upgrades": system_prepare_package_manager_disable_unattended_upgrades,
            "manage_packages": system_prepare_package_manager_manage_packages
        },
        "ntp": {
            "chrony": system_prepare_system_chrony,
            "timesyncd": system_prepare_system_timesyncd
        },
        "system": {
            "setup_selinux": system_prepare_system_setup_selinux,
            "setup_apparmor": system_prepare_system_setup_apparmor,
            "disable_firewalld": system_prepare_system_disable_firewalld,
            "disable_swap": system_prepare_system_disable_swap,
            "modprobe": system_prepare_system_modprobe,
            "sysctl": system_prepare_system_sysctl,
            "audit": {
                "install": system_install_audit,
                "configure": system_prepare_audit,
            }
        },
        "cri": {
            "install": system_cri_install,
            "configure": system_cri_configure
        },
        "thirdparties": system_prepare_thirdparties
    },
    "deploy": {
        "loadbalancer": {
            "haproxy": {
                "install": deploy_loadbalancer_haproxy_install,
                "configure": deploy_loadbalancer_haproxy_configure,
            },
            "keepalived": {
                "install": deploy_loadbalancer_keepalived_install,
                "configure": deploy_loadbalancer_keepalived_configure,
            }
        },
        "kubernetes": {
            "reset": deploy_kubernetes_reset,
            "install": deploy_kubernetes_install,
            "prepull_images": deploy_kubernetes_prepull_images,
            "init": deploy_kubernetes_init,
            "audit": deploy_kubernetes_audit,
        },
        "coredns": deploy_coredns,
        "plugins": deploy_plugins,
        "accounts": deploy_accounts
    },
    "overview": overview
})

cumulative_points = {

    # Reboot and verify that the most crucial system settings are applied on boot.
    # This is done before `prepare.system.audit`.
    system.reboot_nodes: [
        "prepare.system.audit"
    ],
    system.verify_system: [
        "prepare.system.audit"
    ],
    # Some checks can be done only at the end when the necessary services are configured.
    summary.exec_delayed: [
        flow.END_OF_TASKS
    ]
}


def run_tasks(res: DynamicResources, tasks_filter: List[str] = None) -> None:
    InstallAction(tasks_filter).run(res)


class InstallAction(flow.TasksAction):
    def __init__(self, tasks_filter: List[str] = None) -> None:
        super().__init__('install', tasks, cumulative_points=cumulative_points, tasks_filter=tasks_filter)
        self.target_version = "not supported"

    def cluster(self, resources: DynamicResources) -> KubernetesCluster:
        cluster = super().cluster(resources)
        self.target_version = kubernetes.get_kubernetes_version(cluster.inventory)
        return cluster


def create_context(cli_arguments: List[str] = None) -> dict:
    cli_help = '''
    Script for installing Kubernetes cluster.

    How to use:

    '''

    parser = flow.new_tasks_flow_parser(cli_help, tasks=tasks)
    context = flow.create_context(parser, cli_arguments, procedure='install')
    return context


def main(cli_arguments: List[str] = None) -> None:
    context = create_context(cli_arguments)
    install = InstallAction()
    flow_ = flow.ActionsFlow([install])
    result = flow_.run_flow(context)

    kubernetes.verify_supported_version(install.target_version, result.logger)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from collections import OrderedDict
from typing import List, Callable, Dict
import uuid
from kubemarine import kubernetes, plugins, admission, jinja
from kubemarine.core import flow, log, resources as res
from kubemarine.core import utils
from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage
from kubemarine.core.resources import DynamicResources
from kubemarine.kubernetes import components
from kubemarine.procedures import install


def cleanup_tmp_dir(cluster: KubernetesCluster) -> None:
    # Clean up kubernetes tmp dir, where backup files from previous upgrades are located
    nodes = cluster.make_group_from_roles(roles=["control-plane", "worker"])
    nodes.sudo("rm -rf $(sudo find /etc/kubernetes/tmp -mindepth 1 -maxdepth 1)")
    cluster.log.debug("Backup files for previous upgrades were cleaned")


def system_prepare_thirdparties(cluster: KubernetesCluster) -> None:
    if not cluster.inventory['services'].get('thirdparties', {}):
        cluster.log.debug("Skipped - no thirdparties defined in config file")
        return

    install.system_prepare_thirdparties(cluster)


def prepull_images(cluster: KubernetesCluster) -> None:
    cluster.log.debug("Prepulling Kubernetes images...")
    upgrade_group = kubernetes.get_group_for_upgrade(cluster)
    upgrade_group.call(kubernetes.images_grouped_prepull)


def kubernetes_upgrade(cluster: KubernetesCluster) -> None:
    initial_kubernetes_version = kubernetes.get_kubernetes_version(cluster.previous_inventory)

    upgrade_group = kubernetes.get_group_for_upgrade(cluster)
    preconfigure_components = []
    preconfigure_functions: Dict[str, Callable[[dict], dict]] = {}
    if (admission.is_pod_security_unconditional(cluster)
            and utils.version_key(initial_kubernetes_version)[0:2] < utils.minor_version_key("v1.28")
            and cluster.inventory['rbac']['pss']['pod-security'] == 'enabled'):

        # Extra args of API server have changed, need to reconfigure the API server.
        # See admission.enrich_inventory()
        # Still, should not reconfigure using generated ConfigMaps from inventory,
        # because the inventory has already incremented kubernetesVersion, but the cluster is not upgraded yet.
        # Instead, change only necessary apiServer args.
        def reconfigure_feature_gates(cluster_config: dict) -> dict:
            feature_gates = cluster.inventory["services"]["kubeadm"]["apiServer"]["extraArgs"].get("feature-gates")
            if feature_gates is not None:
                cluster_config["apiServer"]["extraArgs"]["feature-gates"] = feature_gates
            else:
                del cluster_config["apiServer"]["extraArgs"]["feature-gates"]

            return cluster_config

        preconfigure_components.append('kube-apiserver')
        preconfigure_functions['kubeadm-config'] = reconfigure_feature_gates

    if (kubernetes.components.kube_proxy_overwrites_higher_system_values(cluster)
            and utils.version_key(initial_kubernetes_version)[0:2] < utils.minor_version_key("v1.29")):

        # Defaults of KubeProxyConfiguration have changed.
        # See kubernetes.enrich_kube_proxy()
        def edit_kube_proxy_conntrack_min(kube_proxy_cm: dict) -> dict:
            expected_conntrack: dict = cluster.inventory['services']['kubeadm_kube-proxy']['conntrack']
            if 'min' not in expected_conntrack:
                return kube_proxy_cm

            actual_conntrack = kube_proxy_cm['conntrack']
            if expected_conntrack['min'] != actual_conntrack.get('min'):
                actual_conntrack['min'] = expected_conntrack['min']

            return kube_proxy_cm

        preconfigure_components.append('kube-proxy')
        preconfigure_functions['kube-proxy'] = edit_kube_proxy_conntrack_min

    if preconfigure_components:
        upgrade_group.call(kubernetes.components.reconfigure_components,
                           components=preconfigure_components, edit_functions=preconfigure_functions)

    drain_timeout = cluster.procedure_inventory.get('drain_timeout')
    grace_period = cluster.procedure_inventory.get('grace_period')
    disable_eviction = cluster.procedure_inventory.get("disable-eviction", True)
    drain_kwargs = {
        'disable_eviction': disable_eviction, 'drain_timeout': drain_timeout, 'grace_period': grace_period
    }

    kubernetes.upgrade_first_control_plane(upgrade_group, cluster, **drain_kwargs)
    first_control_plane = cluster.nodes['control-plane'].get_first_member()

    # After first control-plane upgrade is finished we may loose our CoreDNS changes.
    # Thus, we need to re-apply our CoreDNS changes immediately after first control-plane upgrade.
    install.deploy_coredns(cluster)

    # In some versions, kubeadm reverts resolvConf to the default during `upgrade apply`
    # Remove default resolvConf from kubelet-config ConfigMap for debian OS family
    first_control_plane.call(components.patch_kubelet_configmap)

    kubernetes.upgrade_other_control_planes(upgrade_group, cluster, **drain_kwargs)

    if cluster.nodes.get('worker', []):
        kubernetes.upgrade_workers(upgrade_group, cluster, **drain_kwargs)

    kubernetes_cleanup_nodes_versions(cluster)


def kubernetes_cleanup_nodes_versions(cluster: KubernetesCluster) -> None:
    if not cluster.context.get('cached_nodes_versions_cleaned', False):
        cluster.log.verbose('Cached nodes versions required')
        cluster.nodes['control-plane'].get_first_member().sudo('rm -f /etc/kubernetes/nodes-k8s-versions.txt')
        cluster.context['cached_nodes_versions_cleaned'] = True
    else:
        cluster.log.verbose('Cached nodes versions already cleaned')


def upgrade_packages(cluster: KubernetesCluster) -> None:
    upgrade_version = kubernetes.get_procedure_upgrade_version(cluster)

    packages = cluster.procedure_inventory.get(upgrade_version, {}).get("packages", {})
    if packages.get("install") is not None or packages.get("upgrade") is not None or packages.get("remove") is not None:
        install.manage_custom_packages(cluster.nodes['all'])


def upgrade_plugins(cluster: KubernetesCluster) -> None:
    upgrade_version = kubernetes.get_procedure_upgrade_version(cluster)

    # upgrade_candidates is a source of upgradeable plugins, not list of plugins to upgrade.
    # Some plugins from upgrade_candidates will not be upgraded, because they have "install: false"
    upgrade_candidates = {}
    for plugin, plugin_item in cluster.inventory["plugins"].items():
        # Both OOB and custom plugins can have templates dependent on the inventory.
        # By default, let's do not re-install plugins if inventory does not change for them.
        #
        # Still there should be an ability to force re-install them with the same target inventory configuration,
        # using just empty spec in the procedure inventory.
        #
        # This requirement makes it impossible to turn off re-installation of OOB plugins,
        # if compatibility map changes, but target inventory does not (e.g. if all images are redefined).
        if (plugin in cluster.procedure_inventory.get(upgrade_version, {}).get("plugins", {})
                or cluster.previous_inventory["plugins"].get(plugin) != plugin_item):
            upgrade_candidates[plugin] = cluster.inventory["plugins"][plugin]

    plugins.install(cluster, upgrade_candidates)


def release_calico_leaked_ips(cluster: KubernetesCluster) -> None:
    """
    Sometimes IPs can stay in Calico IPAM despite not being used. 
    You can check this by running "calicoctl ipam check --show-problem-ips".
    Those IPs are cleaned by Calico garbage collector, but it can take about 20 minutes.
    This task releases problem IPs with force.
    """
    # Identify the first control plane node
    first_control_plane = cluster.nodes['control-plane'].get_first_member()
    cluster.log.debug("Getting leaked IPs...")

    # Generate a unique report name
    random_report_name = "/tmp/%s.json" % uuid.uuid4().hex
    try:
        # Run calicoctl ipam check and save the results
        first_control_plane.sudo(
            f"calicoctl ipam check --show-problem-ips -o {random_report_name} "
            "| grep 'leaked' || true", hide=False
        )
        cluster.log.debug(f"IPAM check completed and results saved to {random_report_name}")

        # Release the leaked IPs
        first_control_plane.sudo(
            f"calicoctl ipam release --from-report={random_report_name} --force", 
            hide=False
        )
    finally:
        # Clean up the temporary report file
        first_control_plane.sudo(f"rm {random_report_name}", hide=False)
        cluster.log.debug(f"Cleaned up report file: {random_report_name}")


tasks = OrderedDict({
    "cleanup_tmp_dir": cleanup_tmp_dir,
    "verify_upgrade_versions": kubernetes.verify_upgrade_versions,
    "thirdparties": system_prepare_thirdparties,
    "prepull_images": prepull_images,
    "kubernetes": kubernetes_upgrade,
    "kubernetes_cleanup": kubernetes_cleanup_nodes_versions,
    "packages": upgrade_packages,
    "plugins": upgrade_plugins,
    "release_calico_leaked_ips": release_calico_leaked_ips,  # Added here
    "overview": install.overview
})


class UpgradeFlow(flow.Flow):
    def __init__(self) -> None:
        self.target_version = "not supported"

    def _run(self, resources: DynamicResources) -> None:
        logger = resources.logger()

        previous_version = kubernetes.get_kubernetes_version(resources.inventory())
        upgrade_plan = resources.procedure_inventory().get('upgrade_plan')
        if not upgrade_plan:
            raise Exception('Upgrade plan is not specified in procedure')
        upgrade_plan = verify_upgrade_plan(previous_version, upgrade_plan, logger)

        args = resources.context['execution_arguments']
        if (args['tasks'] or args['exclude']) and len(upgrade_plan) > 1:
            raise Exception("Usage of '--tasks' and '--exclude' is not allowed when upgrading to more than one version")

        # todo inventory is preserved few times, probably need to preserve it once instead.
        actions = [UpgradeAction(version, i) for i, version in enumerate(upgrade_plan)]
        flow.run_actions(resources, actions)
        self.target_version = actions[-1].upgrade_version


class UpgradeAction(flow.TasksAction):
    def __init__(self, upgrade_version: str, upgrade_step: int) -> None:
        super().__init__(f'upgrade step {upgrade_step + 1}', tasks,
                         recreate_inventory=True)
        self.upgrade_version = upgrade_version
        self.upgrade_step = upgrade_step

        if upgrade_step > 0:
            del self.tasks['cleanup_tmp_dir']

    def cluster(self, res: DynamicResources) -> KubernetesCluster:
        # Make sure to enrich at DEFAULT stage without impact from changed context
        res.cluster(EnrichmentStage.DEFAULT)

        context = res.context
        context['upgrade_step'] = self.upgrade_step
        res.reset_cluster(EnrichmentStage.DEFAULT)
        # This starts PROCEDURE enrichment
        cluster = super().cluster(res)

        # New version is enriched and compiled
        self.upgrade_version = kubernetes.get_kubernetes_version(cluster.inventory)
        self.identifier = f'upgrade to {self.upgrade_version}'

        return cluster

    def run(self, resources: res.DynamicResources) -> None:
        super().run(resources)
        # Change context back, but do not DynamicResources.reset_cluster().
        # Once the inventory is recreated, the reset context will be picked up during new cluster initialization.
        del resources.context['upgrade_step']

    def prepare_context(self, context: dict) -> None:
        context['dump_subdir'] = 'upgrade' if jinja.is_template(self.upgrade_version) else self.upgrade_version


def create_context(cli_arguments: List[str] = None) -> dict:
    cli_help = '''
    Script for automated upgrade of the entire Kubernetes cluster to a new version.

    How to use:

    '''

    parser = flow.new_procedure_parser(cli_help, tasks=tasks)

    context = flow.create_context(parser, cli_arguments, procedure='upgrade')
    return context


def main(cli_arguments: List[str] = None) -> None:
    context = create_context(cli_arguments)
    flow_ = UpgradeFlow()
    result = flow_.run_flow(context)

    kubernetes.verify_supported_version(flow_.target_version, result.logger)


def verify_upgrade_plan(previous_version: str, upgrade_plan: List[str], logger: log.EnhancedLogger) -> List[str]:
    # This validates upgrade possibility only partially, and lacks of CRI, PSP/PSS or other validations.
    # It could be better to run enrichment for all the versions before even one upgrade.
    is_any_template = False
    for version in itertools.chain((previous_version,), upgrade_plan):
        if not jinja.is_template(version):
            kubernetes.verify_allowed_version(version)
        else:
            is_any_template = True

    if not is_any_template:
        pv = previous_version
        for version in upgrade_plan:
            kubernetes.test_version_upgrade_possible(pv, version)
            pv = version

        logger.debug(f"Loaded upgrade plan: current ({previous_version}) ⭢ {' ⭢ '.join(upgrade_plan)}")
    else:
        logger.debug(f"Cannot early validate upgrade versions for jinja templates")

    return upgrade_plan


if __name__ == '__main__':
    main()
